/*
    ArchiveSMTP Milter v1.1.b1
    --------------------------
	Copyright (c) 2008, 2009, 2010 Dancing Fortune Software.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
 
	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 
	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */

#include "ArchiveSMTP.h"

/* Global Variables */
Rule		*firstRule;			/* Start of rule list */
int			verboseLevel;		/* Verbosity level */
bool		alwaysAccept;		/* Accept policy (default to reject on error) */
bool		addHeadersSupported;
bool		weUseAddHeaders;

/* Filter Registration Data, Name, API version, call backs to register for */
struct smfiDesc smfilter =
{
	"ArchiveSMTP",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDRCPT | SMFIF_ADDHDRS,  /* flags */
	mlfi_connect,   /* connection info filter */
	mlfi_helo,      /* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	mlfi_envrcpt,   /* envelope recipient filter */
	mlfi_header,	/* header filter */
	mlfi_eoh,       /* end of header */
	mlfi_body,      /* body block filter */
	mlfi_eom,       /* end of message */
	mlfi_abort,     /* message aborted */
	mlfi_close,     /* connection cleanup */
	mlfi_unknown,   /* unknown SMTP commands */
	mlfi_data,      /* DATA command */
	mlfi_negotiate  /* Once, at the start of each SMTP connection */
};

/* mlfi_envfrom:
 Called on "mail from" cmd from client, sets up private data block for
 this mail session and determines local user name of sender.
 */
sfsistat mlfi_envfrom(SMFICTX *ctx, char **argv)
{
    /* Set up our private data block for this session */
    mlfiPriv_t *priv = malloc(sizeof(mlfiPriv_t));
    
    /* Did we allocate memory? */
    if (priv == NULL) {
        MemErrorPrint("mlfi_envfrom", "private data", 0);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL; /* Reject mail with temporary error */
    }
    
    /* Set initial pointers to NULL */
	priv->mlfi_toAddr_list = NULL;
    priv->mlfi_header_list = NULL;
    priv->mlfi_body_chunks = NULL;
    priv->mlfi_username = NULL;
    priv->mlfi_useraddr = NULL;
	priv->mlfi_rcptaddr = NULL;
    priv->mlfi_bytes = 0;
	priv->mlfi_matched = FALSE;
    
    /* Associate data block with this messages context */
	smfi_setpriv(ctx, priv);
    
    /* Get the username of this mail client and copy to our struct */
    char *username;
    
    username = smfi_getsymval(ctx, "{auth_authen}");
    if (username != NULL) {
		/* Username was set, store it */
		char *usercopy = malloc(sizeof(char)*(strlen(username)+1));
		strcpy(usercopy, username);
		priv->mlfi_username = usercopy;
	}
	
	/* Store from address */
	char *addrcopy = ParseMailAddr(argv[0]);
	if (addrcopy == NULL) {
		/* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_envfrom", "sender address", 0);
		return mlfi_cleanup(ctx, FALSE);
	}
    priv->mlfi_useraddr = addrcopy;
    
    /* To help with file locking later, calculate the size of the message so far,
	 it should start with the from address plus an extra 9 characters for the word
	 "from" then two spaces followed by the opening < and closing > */
	priv->mlfi_bytes += ((strlen(addrcopy) + 9) * sizeof(char));
    
	/* Continue processing */
	return SMFIS_CONTINUE;
}

/* mlfi_envrcpt:
 Called on "rcpt to" cmd from client. We will collect this data later with the
 rest of the headers, only here for debugging purposes.
 */
sfsistat mlfi_envrcpt(SMFICTX *ctx, char **argv)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
        MemErrorPrint("mlfi_envrcpt", NULL, 1);
        if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
    }
	
	/* Add the recipient address to the list */
	toAddrList_t *address = NULL;
	if (argv[0] != NULL) {
		address = malloc(sizeof(toAddrList_t));
		if (address == NULL) {
			/* Reject mail with temporary failure if we can't allocate memory */
			MemErrorPrint("mlfi_envrcpt", "recipient structure", 0);
			return mlfi_cleanup(ctx, FALSE);
		}
		
		address->address = ParseMailAddr(argv[0]);
		if (address->address == NULL) {
			/* Reject mail with temporary failure if we can't allocate memory */
			MemErrorPrint("mlfi_envrcpt", "recipient address", 0);
			return mlfi_cleanup(ctx, FALSE);
		}
		address->matched = FALSE;
		address->next = NULL;
	}
	
	/* Add this address struct to list */
    if (priv->mlfi_toAddr_list == NULL) {
        priv->mlfi_toAddr_list = address;
	}
    else {
        /* Iterate through list until we find the last item */
        toAddrList_t *ptr = priv->mlfi_toAddr_list;
        while (ptr->next != NULL) {
            ptr = ptr->next;
		}
        ptr->next = address;
    }
	
    return SMFIS_CONTINUE;
}

/* mlfi_header:
 Called for each header from client, copies the headers into our private
 data block as a linked list with header field name and value. Note: the
 trailing CR/LF (\r\n) is removed from the value.
 */
sfsistat mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_header", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
    
    /* Allocate and fill header stucture memory */
    if (headerf == NULL) {
		/* Is this possible? Check for it just in case outlook decides it is */
        fprintf(stderr, "ArchiveSMTP(mlfi_header): WARNING - Header name is not set, ignoring this header.\n");
        return SMFIS_CONTINUE;
    }
    
	headerList_t *header = malloc(sizeof(headerList_t));
    header->headerName = malloc(sizeof(char)*(strlen(headerf)+1));
    if (header == NULL || header->headerName == NULL) {
		/* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_header", "header name", 0);
        return mlfi_cleanup(ctx, FALSE);
    }
	
    strcpy(header->headerName, headerf);
	
	/* Check there is a header value (outlook sometimes puts headers with no value),
	 if there is a value allocate memory and store it. */
    if (headerv != NULL) {
        header->headerValue = malloc(sizeof(char)*(strlen(headerv)+1));
        if (header->headerValue == NULL) {
            /* Reject mail with temporary failure if we can't allocate memory */
			MemErrorPrint("mlfi_header", "header value", 0);
			return mlfi_cleanup(ctx, FALSE);
        }
        strcpy(header->headerValue, headerv);
    }
	
    header->next = NULL;
    
    /* Add this header struct to list */
    if (priv->mlfi_header_list == NULL) {
        priv->mlfi_header_list = header;
	}
    else {
        /* Iterate through list until we find the last item */
        headerList_t *ptr = priv->mlfi_header_list;
        while (ptr->next != NULL) {
            ptr = ptr->next;
		}
        ptr->next = header;
    }
    
    /* Add field sizes to total message size (plus 3 for ": " and \n) */
    priv->mlfi_bytes += ((strlen(header->headerName) + strlen(header->headerValue)
                         + 3) * sizeof(char));
    
	/* Continue processing */
	return SMFIS_CONTINUE;
}

/* mlfi_body:
 Called for each body chunk from client, can be called multiple times and
 data is not c string formatted. Copies body chunks into private data struct
 as a linked list. 
 */
sfsistat mlfi_body(SMFICTX *ctx, unsigned char *bodyp, size_t bodylen)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_body", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
    
    /* Allocate and fill struct for this body chunk */
    bodyChunks_t *body = malloc(sizeof(bodyChunks_t));
    if (body == NULL) {
        /* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_body", "body structure", 0);
        return mlfi_cleanup(ctx, FALSE);
    }
    body->body = malloc(bodylen);
    if (body->body == NULL) {
        /* Reject mail with temporary failure if we can't allocate memory */
		MemErrorPrint("mlfi_body", "body data", 0);
        return mlfi_cleanup(ctx, FALSE);
    }
	
	/* Body is NOT a string and can contain \0, so we must use memcpy */
    memcpy(body->body, bodyp, bodylen);
    body->size = bodylen;
    body->next = NULL;
    
    /* Add this body chunk to the list */
    if (priv->mlfi_body_chunks == NULL) {
        priv->mlfi_body_chunks = body;
	}
    else {
        /* Iterate through list to find last item */
        bodyChunks_t *ptr = priv->mlfi_body_chunks;
        while (ptr->next != NULL) {
            ptr = ptr->next;
		}
        ptr->next = body;
    }
    
    /* Add body size to total size */
    priv->mlfi_bytes += bodylen;
    
	/* Continue processing */
	return SMFIS_CONTINUE;
}

/* mlfi_eom:
 Called at the end of a message from client, at this point we should have got
 all the message data, so we can write it to the correct user owned archive.
 */
sfsistat mlfi_eom(SMFICTX *ctx)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_eom", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
	
	/* Generate timestamp first, so that all output has the same stamp */
	time_t stamp = time(NULL);
	/* Add size of time stamp to total size (plus 3 for eom padding and blank line after headers) */
	priv->mlfi_bytes += ((strlen(ctime(&stamp)) + 3) * sizeof(char));
	
	/* Now we must evaluate the collected data against our rule set to see if
	 and where to store this e-mail or what other actions to take. */
	Rule *itr = firstRule;
	Action *actionItr;
	bool result = TRUE;
	char *subbed = NULL;
	while (itr) 
	{
		if (EvaluateRule(itr, priv))
		{
			/* Indicate it has matched something */
			priv->mlfi_matched = TRUE;
			
			/* Perform all actions */
			actionItr = itr->action;
			while (actionItr) {
				if (verboseLevel > 0) {
					if (!actionItr->allTo)
						subbed = SubTokens(actionItr, priv);
					else
						subbed = "Each matched to address:";
				}
				switch (actionItr->type) {
					case kActionStoreIn:
						VerbosePrint(1, "Action: Store In '%s' ", subbed);
						if (!ActionStoreIn(actionItr, &stamp, priv)) {
							result = FALSE;
							VerbosePrint(1, "- %s ", "Failed.");
						} else
							VerbosePrint(1, "- %s ", "Success.");
						break;
					case kActionAddHeader:
						VerbosePrint(1, "Action: Add Header '%s' ", subbed);
						if (!ActionAddHeader(actionItr, ctx, priv)) {
							result = FALSE;
							VerbosePrint(1, "- %s ", "Failed.");
						} else
							VerbosePrint(1, "- %s ", "Success.");
						break;
					case kActionPipe:
						VerbosePrint(1, "Action: Pipe '%s' ", subbed);
						if (!ActionPipe(actionItr, &stamp, priv)) {
							result = FALSE;
							VerbosePrint(1, "- %s ", "Failed.");
						} else
							VerbosePrint(1, "- %s ", "Success.");
						break;
					default:
						break;
				}
				if (verboseLevel > 0) {
					if (!actionItr->allTo)
						free(subbed);
					else
						subbed = NULL;
				}
				actionItr = actionItr->next;
			}
			/* Should we halt on this rule matching? */
			if (itr->halt) {
				VerbosePrint(1, "%s", "Halting.\n");
				break;
			}
		}
		itr = itr->next;
		VerbosePrint(1, "%s", "\n");
	}
    
    /* Call clean-up function */
	return mlfi_cleanup(ctx, result);
}

/* mlfi_abort:
 Called when for what ever reason a message is aborted before being sent, we
 will just discard any aborted mails.
 */
sfsistat mlfi_abort(SMFICTX *ctx)
{
    /* Call clean-up function, nothing more to be done */
	return mlfi_cleanup(ctx, TRUE);
}

/* mlfi_cleanup:
 Called internally at end of processing to free memory used for this context
 */
sfsistat mlfi_cleanup(SMFICTX *ctx, bool okay)
{
    /* Grab our private session data */
    mlfiPriv_t *priv = MLFIPRIV;
    if (priv == NULL) {
		/* This shouldn't ever happen, but to be safe we test for it */
		MemErrorPrint("mlfi_cleanup", NULL, 1);
		if (alwaysAccept)
			return SMFIS_CONTINUE;
		else
			return SMFIS_TEMPFAIL;
	}
	
	/* Free recipient list */
	toAddrList_t *ptrT = priv->mlfi_toAddr_list;
	while (ptrT != NULL) {
		priv->mlfi_toAddr_list = ptrT->next;
		free(ptrT->address);
		free(ptrT);
		ptrT = priv->mlfi_toAddr_list;
	}
    
    /* Free header list */
    headerList_t *ptrH = priv->mlfi_header_list;
    while (ptrH != NULL) {
        // Free the header struct
        priv->mlfi_header_list = ptrH->next;
        if (ptrH->headerName != NULL)
            free(ptrH->headerName);
        if (ptrH->headerValue != NULL)
            free(ptrH->headerValue);
        free(ptrH);
        ptrH = priv->mlfi_header_list;
    }
    
    /* Free body chunks */
    bodyChunks_t *ptrB = priv->mlfi_body_chunks;
    while (ptrB != NULL) {
        priv->mlfi_body_chunks = ptrB->next;
        free(ptrB->body);
        free(ptrB);
        ptrB = priv->mlfi_body_chunks;
    }
    
    /* Milter API docs say to set ptr to null before terminating */
    free(priv);
    smfi_setpriv(ctx, NULL);
    
	if (!okay && !alwaysAccept) {
		/* We failed to collect this mail, tell the mta to reject it with a
		 temporary failure message so they will try again. */
		return SMFIS_TEMPFAIL;
	}
	/* Let mta continue processing */
	return SMFIS_CONTINUE;
}

/* mlfi_negotiate: Tells the MTA what our conditions are and stuff like that. */
sfsistat mlfi_negotiate 
(SMFICTX *ctx,
 unsigned long f0,
 unsigned long f1,
 unsigned long f2,
 unsigned long f3,
 unsigned long *pf0,
 unsigned long *pf1,
 unsigned long *pf2,
 unsigned long *pf3)
{
	const unsigned long ourDesiredActions = SMFIP_NOCONNECT | SMFIP_NOHELO 
	| SMFIP_NOEOH | SMFIP_NOUNKNOWN
	| SMFIP_NODATA;
	/* Check the MTA supports our required actions.
	 Since we have no required actions, this is easy. */
	if ((f0 & smfilter.xxfi_flags) != smfilter.xxfi_flags)
	{
		/* The actions we need are NOT supported. We have no required
		 actions right now, so this doesn't really matter, but if
		 we did have some we could probably use it to enable/disable
		 features on the fly and only error if the config uses them. */
		addHeadersSupported = FALSE;
		*pf0 = 0x0;
		if (weUseAddHeaders)
			fprintf(stderr, "WARNING: the config file includes Add Headers actions, but the MTA **does not support** adding headers!!\nAdd Headers actions will be ignored!\n");
	}
	else {
		addHeadersSupported = TRUE;
		*pf0 = SMFIF_ADDHDRS | SMFIF_ADDRCPT;
	}
	
	*pf1 = f1 & ourDesiredActions; /* only send actions the MTA supports */
	*pf2 = 0x0; /* reserved by API; must be 0 */
	*pf3 = 0x0; /* reserved by API; must be 0 */
	
	return SMFIS_CONTINUE;
}

/* Unused Callbacks: Defined only so we can negotiate with an MTA that doesn't
 support ignoring these callbacks. */
sfsistat mlfi_connect (SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr) { return SMFIS_CONTINUE; }
sfsistat mlfi_helo (SMFICTX *ctx, char *helohost) { return SMFIS_CONTINUE; }
sfsistat mlfi_data (SMFICTX *ctx) { return SMFIS_CONTINUE; }
sfsistat mlfi_unknown (SMFICTX *ctx, const char *arg) { return SMFIS_CONTINUE; }
sfsistat mlfi_eoh (SMFICTX *ctx) { return SMFIS_CONTINUE; }
sfsistat mlfi_close (SMFICTX *ctx) { return SMFIS_CONTINUE; }

/* ParseMailAddr: Removes the leading '<' and trailing '>' from a mail address
 and returns a newly allocted string containing just the address. */
char* ParseMailAddr(const char *addr)
{
	char *addrcopy;
	int len = strlen(addr);
	int i, j;
	
	/* Check that the address actually starts with '<' and ends with '>' */
	if (addr[0] == '<' && addr[(len-1)] == '>') {
		addrcopy = malloc(sizeof(char)*(len+1 - 2));
		if (addrcopy == NULL) {
			return (NULL);
		}
		i = 1;
		len = (len-2);
		addrcopy[len] = '\0';
	} else {
		addrcopy = malloc(sizeof(char)*(len+1));
		if (addrcopy == NULL) {
			return (NULL);
		}
		i = 0;
		len = (len+1);
	}
	
	/* Copy characters into the new string, changing them all the lower case */
	for (j = 0; j < len; i++, j++) {
		addrcopy[j] = tolower(addr[i]);
	}
	
	return (addrcopy);
}

/* MemErrorPrint: Prints a message to stderr about a failed memory allocation */
void MemErrorPrint(const char *func, const char *obj, int type)
{
	char *action;
	char action1[] = "Mail accepted despite error";
	char action2[] = "Mail returned to sender";
	
	if (alwaysAccept)
		action = action1; 
	else
		action = action2;
	
	switch (type) {
		case 0:
			fprintf(stderr, "ArchiveSMTP(%s): ERROR - Could not allocate memory for %s.\
					perhaps system or process memory limits have been reached? %s.\n",
					func, obj, action);
			break;
		case 1:
			fprintf(stderr, "ArchiveSMTP(%s): ERROR - Private data was not found!\
					Memory has been corrupted? %s.\n", func, action);
			break;
		case 2:
			fprintf(stderr, "ArchiveSMTP(%s): ERROR - %s (errno: %d:%s). %s.\n", func, obj, errno, strerror(errno), action);
			break;
		default:
			break;
	}
}

/* EatWhiteSpace: Moves forward in a file, eating white space characters until it
 finds non-white space, returning the first non-white space character it finds. */
char EatWhitespace(FILE* file)
{
	char c;
	int inComment = 0;
	while (1)
	{
		switch (c = fgetc(file))
		{
			case ' ':
			case '\t':
			case '\v':
			case '\r':
				break;
			case '\n':
				/* End comment if in one */
				if (inComment)
					inComment = FALSE;
				break;
			case '#':
				/* Also eat comments */
				if (!inComment)
					inComment = TRUE;
				break;
			default:
				/* Return the character if its not part of a comment */
				if (!inComment)
					return c;
				break;
		}
	}
}

/* EatWhiteAndReadString: Moves forward in a file, eating white space characters
 until it finds non-white space. The characters preceeding the white space until
 the next white space are read into a buffer and returned as a newly allocted
 string. Quote characters and backslash escapes are honoured, the buffer is limited
 to 256 bytes. */
char* EatWhiteAndReadString(FILE* file)
{
	char	buffer[256];
	char	c;
	char	*theStr;
	int		parse = TRUE;
	int		quote = FALSE;
	int		i = 0;
	
	/* Eat any leading whitespace */
	c = EatWhitespace(file);
	/* Parse file for a string, including quotes and escapes */
	while(parse) {
		switch(c) {
			case '\"':				/* Open or Close quotes */
				if (!quote)
					quote = TRUE;
				else
					quote = FALSE;
				i--;
				break;
			case '\\':				/* Escape - Read next character literally */
				buffer[i] = fgetc(file);
				break;
			case '\0':				/* this is always an error */
				fprintf(stderr, "Unexpected end-of-file while trying to parse string.\n");
				exit(EX_SOFTWARE);
				break;
			case ';':				/* End rule actions */
			case ',':				/* Action separator */
			case ')':				/* Close bracket */
			case '}':				/* Close rule */
				if (!quote)
					ungetc(c, file);
			case ' ':				/* White space */
			case '\t':
			case '\v':
			case '\r':
			case '\n':
				if (!quote) {
					parse = FALSE;
					buffer[i] = '\0';
					break;
				}
			default:				/* All other characters */
				buffer[i] = c;
				break;
		}
		/* Get next character if still parsing */
		if (parse) {
			c = fgetc(file);
			i++;
			
			/* Explode on buffer overflow */
			if (i >= 256) {
				fprintf(stderr, "Buffer overflow! Perhaps the configuration file is missing a closing quote?\n");
				exit(EX_SOFTWARE);
			}
		}
	}
	
	/* Allocate and return the string */
	theStr = malloc(sizeof(char)*(i + 1));
	strcpy(theStr, buffer);
	
	return (theStr);
}

/* StringToLower: Goes through a null terminated string, turning any uppercase
 characters to lowercase. The original string is modified. */
void StringToLower(char *str)
{
	int i;
	for (i = 0; i < (strlen(str)); i++) {
		str[i] = tolower(str[i]);
	}
}

/*  Returns TRUE if the passed string is next in the file and eats it.
	Returns FALSE otherwise. */
int CheckStringIsNextInFile(FILE* file, const char* string, const char* errorMessage)
{
	int i;
	char c;
	for (i = 0; string[i]; i++)
	{
		c = tolower(fgetc(file));
		if (c != string[i])
		{
			fprintf(stderr, errorMessage);
			exit(EX_SOFTWARE);
		}
		/*	return FALSE; */
	}
	return TRUE;
}

void RuleStartedParsingState(FILE* file, Rule* theRule)
{
	/*
			This parsing state is entered after we encounter an opening bracket,
		signifying that a rule definition has started.
		
		Here is a reference example of a rule:
		
		{	Sender equals Fred or From begins fred	} copy to "/Logs/"
	 */
	char	input;
	bool	parse;
	Action	*theAction;
	
	/* Read the details of the condition. This will stop when we reach the } */
	theRule->root = ParseCondition(file);
	/* so at this point we just found a closing brace. Let's eat the whitespace. */
	input = EatWhitespace(file);
	
	/* The next part should be the instruction for what to do with the mail. This
		is probably "store in" at this point, followed by a location. So check the
		first character; if it's not Ss, explode, and if it is then check for the
		rest of tore in. If it's not there then error out. */
	
	parse = TRUE;
	
	while (parse)
	{
		switch(input) {
			case 'S':
			case 's':
				CheckStringIsNextInFile(file, "tore in", "Parsing error - expected Store In\n");
				/* We now need to eat the whitespace, then read the rest of the filename that should
				 be here. */
				theAction = GetNextActionSlot(theRule);
				theAction->type = kActionStoreIn;
				theAction->allTo = FALSE;
				theAction->addrItr = NULL;
				theAction->args = PreProcessTokens(EatWhiteAndReadString(file), theRule, theAction);
				theAction->pathLen = GetPreTokenLen(theAction->args);
				theAction->next = NULL;
				break;
			case 'P':
			case 'p':
				CheckStringIsNextInFile(file, "ipe", "Parsing error - expected Pipe\n");
				theAction = GetNextActionSlot(theRule);
				theAction->type = kActionPipe;
				theAction->allTo = FALSE;
				theAction->addrItr = NULL;
				theAction->args = PreProcessTokens(EatWhiteAndReadString(file), theRule, theAction);
				theAction->pathLen = GetPreTokenLen(theAction->args);
				theAction->next = NULL;
				break;
			case 'A':
			case 'a':
				CheckStringIsNextInFile(file, "dd header", "Parsing error - expected Add Header\n");
				theAction = GetNextActionSlot(theRule);
				theAction->type = kActionAddHeader;
				theAction->allTo = FALSE;
				theAction->addrItr = NULL;
				theAction->args = PreProcessTokens(EatWhiteAndReadString(file), theRule, theAction);
				theAction->pathLen = GetPreTokenLen(theAction->args);
				theAction->next = NULL;
				weUseAddHeaders = TRUE;
				break;
			case 'H':
			case 'h':
				CheckStringIsNextInFile(file, "alt", "Parsing error - expected Halt\n");
				theRule->halt = TRUE;
				break;				
			default:
				/* Explode - No action found after closing bracket */
				fprintf(stderr, "Parsing error - expected at least one action\n");
				exit(EX_SOFTWARE);
				break;
		}
		/* check for end of rule */
		input = EatWhitespace(file);
		switch (input)
		{
			case ',':
				/* this means we want to parse another action, so repeat */
				input = EatWhitespace(file);
				break;
			case ';':
				/* this means end of rule */
				parse = FALSE;
				break;
			default:
				fprintf(stderr, "Parsing error - expected , or ;\nBe aware that the config file format now requires rules to end in \";\", so your config file may be out of date\n");
				exit(EX_SOFTWARE);
				break;
		}
	}
}

void InitialParsingState(FILE* file)
{
	/*
		This is the first parsing state, called immediately after we start 
	 */
	char	input;
	int		parse = 1;
	int		inComment = 0;
	Rule*	newRule;
	
	weUseAddHeaders = FALSE;
	
	while (parse)
	{
		input = fgetc(file);
		switch (input)
		{
			case '\0':
			case EOF:
				/* eof - stop parsing */
				parse = 0;
				break;
			case ' ':
			case '\t':
			case '\v':
			case '\r':
				/* whitespace - ignore */
				break;
			case '\n':
				/* terminate comment if one is open else ignore */
				if (inComment)
					inComment = FALSE;
				break;
			case '#':
				/* Start of a comment - ignore rest of line */
				if (!inComment)
					inComment = TRUE;
				break;
			case '{':
				/* open bracket - start a thingy */
				if (!inComment) {
					newRule = CreateNewRule();
					AddRuleToRuleList(newRule);
					RuleStartedParsingState(file, newRule);
				}
				break;
			default:
				/* found something else - panic unless in a comment */
				if (!inComment) {
					fprintf(stderr, "Parsing error - unexpected character: '%c'\n", input);
					exit(EX_SOFTWARE);
				}
				break;
		}
	}
}

void ReadConfigurationFile(const char* filename)
{
	FILE* file = fopen(filename, "r");
	
	if (!file)
	{
		fprintf(stderr, "Unable to open config file at \"%s\" (errno: %d:%s). Aborting.\n", filename, errno, strerror(errno));
		exit(EX_SOFTWARE);
	}
	
	/* file opened successfully, so we can start parsing. */
	InitialParsingState(file);
	
	fclose(file);
}

/* Usage string, should be used by 'getopt' magic */
static void usage(char *prog)
{
	fprintf(stderr,
            "Usage: %s [-adhnv] [-r pidfile] [-t timeout] [-u username] -f filename -p port\n",
            prog);
}

/* main:
 Archive SMTP Milter - Main (setup, register, pass off to libmilter main), code
 taken mostly as-is from milter example main.
 */
int main (int argc, char * const argv[])
{   
	/* Initialise Globals */
	firstRule = NULL;
	verboseLevel = 0;
	alwaysAccept = TRUE;
	
    /* Vars defined by getopt:
	 * extern char *optarg;
     * extern int optind;
     * extern int optopt;
     * extern int opterr;
     * extern int optreset;
	 */
	
    /* Our argument list */
    const char *args = "adf:hnp:r:t:u:v";
	
	/* Argument Set Flags */
	bool	sockPortSet = FALSE;
	bool	confFileSet = FALSE;
	bool	userNameSet = FALSE;
	bool	testConfig	= FALSE;
	bool	pidFileSet	= FALSE;
	bool	timeOutSet	= FALSE;
	bool	deleteSock	= TRUE;
	
	/* Argument Strings */
	char	*sockPort	= NULL;
	char	*confFile	= NULL;
	char	*userName	= NULL;
	char	*pidFile	= NULL;
	char	*timeOut	= NULL;
    
	/* Process command line options */
	char c;
	while ((c = getopt(argc, argv, args)) != -1)
	{
		switch (c)
		{
			case 'a':	/* Always store mail, reject mail if storage fails */
				alwaysAccept = FALSE;
				break;
			case 'd':	/* Don't delete socket on start up */
				deleteSock = FALSE;
				break;
			case 'f':	/* Configuration file path		- REQUIRED */
				confFileSet = TRUE;
				confFile = optarg;
				break;
			case 'n':	/* Parse & Print config then exit */
				testConfig = TRUE;
				break;
			case 'p':	/* Port or Socket used for IPC	- REQUIRED */
				sockPortSet = TRUE;
				sockPort = optarg;
				break;
			case 'r':	/* File to store process id in */
				pidFileSet = TRUE;
				pidFile = optarg;
				break;
			case 't':	/* Timeout value on MTA comms */
				timeOutSet = TRUE;
				timeOut = optarg;
				break;
			case 'u':	/* Username to run as */
				userNameSet = TRUE;
				userName = optarg;
				break;
			case 'v':	/* Increase verbosity level */
				verboseLevel++;
				break;
			case 'h':	/* Just wants to see usage string */
			default:	/* Unkown argument, print usage string and exit */
                usage(argv[0]);
                exit(EX_USAGE);
				break;
		}
	}
	
	/* First action should be to switch user if one was given */
	if (userNameSet) {
		/* Check if they are root, only root can change user */
		if (getuid() != 0) {
			fprintf(stderr, "You must be root to use the: -u option\n");
			exit(EX_USAGE);
		}
		else if (userName == NULL || *userName == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -u: username\n");
			exit(EX_USAGE);
		}
		
		/* Lookup username and attempt to switch uid/gid to that user */
		struct passwd *uid;
		uid = getpwnam(userName);
		if (uid == NULL) { 
			fprintf(stderr, "Unable to switch user to: %s - No such user\n", userName);
			exit(EX_USAGE);
		}
		else if (setgid(uid->pw_gid) != 0 || setuid(uid->pw_uid) != 0) {
			fprintf(stderr, "Unable to switch uid/gid to: %s (errno: %d:%s)\n", userName, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
	}
	
	/* Next we should parse the config file, this file must exist */
	if (confFileSet) {
		if (confFile == NULL || *confFile == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -f: filename\n");
			exit(EX_USAGE);
		}
		
		/* Process configuration file */
		ReadConfigurationFile(confFile);
	} else {
		usage(argv[0]);
		fprintf(stderr, "Missing required argument: -f\n");
		exit(EX_USAGE);
	}
		
	/* If the testConfig option is set, we will just print the rules and exit */
	if (testConfig) {
		Rule *itr = firstRule;
		while (itr) {
			PrintCondition(itr->root);
			/* print the location of the output */
			// Add printing of all actions in list printf(" to \"%s\"\n", itr->storeIn);
			itr = itr->next;
		}
		exit(EX_OK);
	}
	
	/* Configure the timeout for MTA read/writes if one was given */
	if (timeOutSet) {
		if (timeOut == NULL || *timeOut == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -t: timeout\n");
			exit(EX_USAGE);
		}
		
		/* Set the time out */
		if (smfi_settimeout(atoi(timeOut)) == MI_FAILURE)
		{
			fprintf(stderr, "Unable to set timeout to: %s\n", timeOut);
			exit(EX_SOFTWARE);
		}
	}

    /* Register fiter with MTA */
	if (smfi_register(smfilter) == MI_FAILURE)
	{
		fprintf(stderr, "Unable to register with MTA\n");
		exit(EX_UNAVAILABLE);
	}
	
	/* Finally we need to set the port or socket specified */
	if (sockPortSet) {
		if (sockPort == NULL || *sockPort == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -p: port\n");
			exit(EX_USAGE);
		}
		
		/* Set the port/socket to use */
		if (smfi_setconn(sockPort) == MI_FAILURE) {
			fprintf(stderr, "Unable to set port to: %s\n", sockPort);
			exit(EX_SOFTWARE);
		}
		
		/* Open the socket */
		if (smfi_opensocket(deleteSock) == MI_FAILURE) {
			fprintf(stderr, "Unable to open port: %s\n", sockPort);
			exit(EX_SOFTWARE);
		}
	} else {
		usage(argv[0]);
		fprintf(stderr, "Missing required argument: -p\n");
		exit(EX_USAGE);
	}
	
	/* Now that we are up and running, write our pid to file if set */
	int pidFD = 0;
	struct stat pidStat;
	if (pidFileSet) {
		if (pidFile == NULL || *pidFile == '\0') {
			usage(argv[0]);
			fprintf(stderr, "Missing argument to -r: pidfile\n");
			exit(EX_USAGE);
		}
		/* Try to open the pid file or create if it doesn't exist */
		pidFD = open(pidFile, O_RDWR | O_CREAT, 0600);
		if (pidFD == -1) {
			fprintf(stderr, "Unable to write pid to: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
		/* Try to obtain exclusive lock on the whole file */
		if (flock(pidFD, LOCK_EX | LOCK_NB) == -1) {
			if (errno == EWOULDBLOCK) {
				/* File is already locked by another instance of the program */
				char buffer[16];
				int i = read(pidFD, buffer, (sizeof(buffer) - 1));
				if (i == -1) {
					fprintf(stderr, "Program is already running, but pidfile was unreadable (errno: %d:%s)\n", errno, strerror(errno));
				} else {
					buffer[i] = '\0';
					fprintf(stderr, "Program is already running, pid: %s\n", buffer);
				}
			} else {
				fprintf(stderr, "Error, pidfile \"%s\" is not a regular file (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			}
			exit(EX_SOFTWARE);
		}
		/* Record pid file stats and write pid to file */
		if (!ftruncate(pidFD, 0)) {
			char pidstr[16];
			snprintf(pidstr, sizeof(pidstr), "%u", getpid());
			if (pwrite(pidFD, pidstr, strlen(pidstr), 0) != strlen(pidstr)) {
				fprintf(stderr, "Unable to write pid to: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
				exit(EX_SOFTWARE);
			}
		} else {
			fprintf(stderr, "Unable to truncate pidfile: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
		if (fstat(pidFD, &pidStat)) {
			fprintf(stderr, "Unable to fstat pid file: %s (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			exit(EX_SOFTWARE);
		}
	}
    
    /* Pass off to milters main */
    int exitVal = smfi_main();
	int exitErr = errno;
    
    /* Clean up globals before exit */
	Rule *ptr = firstRule, *oldPtr;
	while (ptr) {
		// Delete stuff
		DeleteRule(ptr);
		oldPtr = ptr;
		ptr = ptr->next;
		free(oldPtr);
	}
	
	if (pidFileSet) {
		/* Delete pid file if stats match saved values */
		struct stat pidFileStat;
		stat(pidFile, &pidFileStat);
		if (pidFileStat.st_dev == pidStat.st_dev &&		/* Check device id */
			pidFileStat.st_ino == pidStat.st_ino &&		/* Check inode */
			pidFileStat.st_mode == pidStat.st_mode &&	/* Check access modes */
			pidFileStat.st_uid == pidStat.st_uid &&		/* Check owner */
			pidFileStat.st_gid == pidStat.st_gid &&		/* Check group */
			pidFileStat.st_rdev == pidStat.st_rdev) {	/* Check the file type */
			/* Perhaps all these checks are over the top? There could still be race condition..? */
			if (unlink(pidFile) == -1) {
				fprintf(stderr, "Unable to remove pidfile \"%s\" on exit (errno: %d:%s)\n", pidFile, errno, strerror(errno));
			}
		} else {
			fprintf(stderr, "Aborting removal of pidfile \"%s\" on exit, file has been changed\n", pidFile);
		}
		if (flock(pidFD, LOCK_UN) == -1) {
			fprintf(stderr, "Unable to remove lock on pidfile \"%s\" on exit (errno: %d:%s)\n", pidFile, errno, strerror(errno));
		}
		if (close(pidFD) == -1) {
			fprintf(stderr, "Unable to close pidfile \"%s\" on exit (errno: %d:%s)\n", pidFile, errno, strerror(errno));
		}
	}
    
	if (exitVal != 0)
		fprintf(stderr, "Exiting with error, this usually means there was a problem with the socket (errno: %d:%s)\n", exitErr, strerror(exitErr));
	
    return (exitVal);
}
