/*
    ArchiveSMTP Milter v1.1.b1
    --------------------------
	Copyright (c) 2008, 2009, 2010 Dancing Fortune Software.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
 
	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 
	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <time.h>
#include <unistd.h>

#include <libmilter/mfapi.h>

typedef struct toAddrList_s toAddrList_t;

#include "Action.h"
#include "Test.h"
#include "Condition.h"
#include "Rule.h"

/* Does this system define a type for bool? */
#ifndef bool
#define bool	int
#define TRUE	1
#define FALSE	0
#endif


/* Mail To address list */
struct toAddrList_s
{
	char	*address;			/* The address value */
	bool	matched;			/* True if a rule condition matched it */
	struct	toAddrList_s *next;	/* Link to next address */
};

/* Mail header structure */
typedef struct headerList_s
{
    char    *headerName;		/* The header field name */
    char    *headerValue;		/* The value of the field */
    struct  headerList_s *next;	/* Link to the next header field */
} headerList_t;

/* Mail body data structure */
typedef struct bodyChunks_s
{
    char    *body;				/* Pointer to body chunk data */
    int     size;				/* Size in bytes of body data */
    struct  bodyChunks_s *next;	/* Link to the next chunk of body data */
} bodyChunks_t;

/* Private instance data */
typedef struct mlfiPriv
{
	toAddrList_t	*mlfi_toAddr_list;	/* Linked list of recipient addresses */
	headerList_t    *mlfi_header_list;	/* Linked list of mail headers */
	bodyChunks_t    *mlfi_body_chunks;	/* Linked list of mail body data */
	char            *mlfi_username;		/* User name of mail sender */
    char            *mlfi_useraddr;		/* From address of mail sender */
	char			*mlfi_rcptaddr;		/* First rept address given */
    int             mlfi_bytes;			/* Total size of mail in bytes */
	bool			mlfi_matched;		/* Flaged when a rule matches */
	
	Rule			*currentRule;
} mlfiPriv_t;

/* Convinence Macro to retrive private data instance */
#define MLFIPRIV	((mlfiPriv_t *) smfi_getpriv(ctx))

/* Verbose level checking and printing Macro */
#define VerbosePrint(L, STR, ARG) if (verboseLevel >= L) { printf(STR, ARG); }

/* Global Variables */
extern Rule	*firstRule;			/* Start of rule list */
extern int	verboseLevel;		/* Verbosity level */
extern bool	alwaysAccept;		/* Accept policy (default to reject on error) */

extern bool addHeadersSupported; /* TRUE if the MTA supports adding headers */
extern bool weUseAddHeaders;	 /* TRUE if we have a rule that adds headers */

/*** Function prototypes ***/

/* Milter call-backs */
sfsistat mlfi_connect	(SMFICTX*, char*, _SOCK_ADDR*);
sfsistat mlfi_helo		(SMFICTX*, char*);
sfsistat mlfi_envfrom	(SMFICTX*, char**);
sfsistat mlfi_envrcpt	(SMFICTX*, char**);
sfsistat mlfi_data		(SMFICTX*);
sfsistat mlfi_unknown	(SMFICTX*, const char*);
sfsistat mlfi_header	(SMFICTX*, char*, char*);
sfsistat mlfi_eoh		(SMFICTX*);
sfsistat mlfi_body		(SMFICTX*, unsigned char*, size_t);
sfsistat mlfi_eom		(SMFICTX*);
sfsistat mlfi_abort		(SMFICTX*);
sfsistat mlfi_close		(SMFICTX*);
sfsistat mlfi_cleanup	(SMFICTX*, bool);
sfsistat mlfi_negotiate (SMFICTX*, unsigned long, unsigned long, unsigned long, unsigned long,
						 unsigned long*, unsigned long*, unsigned long *, unsigned long*);

/* Configuration reading & Rule building/deleting */
void ReadConfigurationFile	(const char* filename);
void InitialParsingState	(FILE* file);
void RuleStartedParsingState(FILE* file, Rule* theRule);
Rule* CreateNewRule			();
void AddRuleToRuleList		(Rule* newRule);
void DeleteRule				(Rule* r);
Action* GetNextActionSlot	(Rule* r);
bool ActionStoreIn			(Action* storein, time_t* stamp, mlfiPriv_t* priv);
bool ActionAddHeader		(Action* addheader, SMFICTX* ctx, mlfiPriv_t *priv);
bool ActionPipe				(Action* pipeto, time_t* stamp, mlfiPriv_t* priv);
Condition* ParseCondition	(FILE* file);
void InitialiseCondition	(Condition* cond);
void DeleteCondition		(Condition* c);
Test* ParseTest				(FILE* file, int testField);

/* Rule testing & printing */
int EvaluateRule			(Rule* r, mlfiPriv_t *priv);
int EvaluateCondition		(Condition* cond, mlfiPriv_t *priv);
int EvaluateConditionParameter(Condition* c, Test* t, mlfiPriv_t *priv, bool negate);
int PerformTest				(Test* t, const char* testData);
void PrintCondition			(Condition* c);
void PrintTest				(Test* t);

/* Utility */
Token* PreProcessTokens		(char* args, Rule* r, Action* action);
int GetPreTokenLen			(Token* tHead);
char* SubTokens				(Action* action, mlfiPriv_t *priv);
void PrintMail				(FILE* output, time_t* stamp, mlfiPriv_t *priv);
char* ParseMailAddr			(const char *addr);
void MemErrorPrint			(const char *func, const char *obj, int type);
int CheckStringIsNextInFile	(FILE* file, const char* string, const char* errorMessage);
char EatWhitespace			(FILE* file);
char* EatWhiteAndReadString	(FILE* file);
void StringToLower			(char *str);

