/*
	ArchiveSMTP Milter v1.1.b1
	--------------------------
	Copyright (c) 2008, 2009, 2010 Dancing Fortune Software.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:

	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.

	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */

#include "ArchiveSMTP.h"

void InitialiseCondition(Condition* cond)
{
	cond->t1 = cond->t2 = NULL;
	cond->p1 = cond->p2 = NULL;
	cond->type = kConditionTypeUndefined;
	cond->negate = FALSE;
}

/*
	PrintCondition(Condition*)
	
	A recursive function that prints the contents of the pointed-to
	passed condition, including its children. Also makes use of
	PrintTest(Test*).
*/
void PrintCondition(Condition* c)
{
	printf("(");
	if (c->negate)
		printf("!");
	if (c->p1)
		PrintCondition(c->p1);
	else if (c->t1)
		PrintTest(c->t1);
	if (c->type == kConditionTypeOr)
		printf(" or ");
	else if (c->type == kConditionTypeAnd)
		printf(" and ");
	if (c->p2)
		PrintCondition(c->p2);
	else if (c->t2)
		PrintTest(c->t2);
	printf(")");
}


int EvaluateConditionParameter(Condition* c, Test* t, mlfiPriv_t *priv, bool negate)
{
	bool	matched = FALSE;
	if (t) /* do we have a test in slot? */
	{
		/* yes! */
		toAddrList_t *ptr = priv->mlfi_toAddr_list;
		switch (t->fieldToTest) {
			case kTestSMTPSender:
				return PerformTest(t, priv->mlfi_username);
				break;
			case kTestFromAddress:
				return PerformTest(t, priv->mlfi_useraddr);
				break;
			case kTestToAddress:
				/* There can be multiple to addresses, so test all of them */
				while (ptr != NULL)
				{
					if (PerformTest(t, ptr->address))
					{
						if (!negate)
							ptr->matched = TRUE;
						matched = TRUE;
						if ((priv->currentRule->flags & kRuleFlagUsesTo) && priv->mlfi_rcptaddr == NULL && !negate)
							priv->mlfi_rcptaddr = ptr->address;
						if (!(priv->currentRule->flags & kRuleFlagUsesToAll))
						{
							return TRUE;		/* Return true on the first match */
						}
					}
					else if (negate) {
						ptr->matched = TRUE;
						if ((priv->currentRule->flags & kRuleFlagUsesTo) && priv->mlfi_rcptaddr == NULL)
							priv->mlfi_rcptaddr = ptr->address;
					}
					VerbosePrint(1, "%s ", ",");
					ptr = ptr->next;
				}
				return (matched);
				break;
			case kTestNoMatch:
				VerbosePrint(1, "%s ", "not matched by a previous rule");
				if (priv->mlfi_matched) {
					VerbosePrint(1, "%s", "[false]");
					return (FALSE);
				} else {
					VerbosePrint(1, "%s", "[true]");
					return (TRUE);
				}			
				break;
			case kTestMatchAll:
				VerbosePrint(1, "%s", "match all [true]");
				return (TRUE);				
				break;
			case kTestUndefined:
			default:
				return (FALSE);
				break;
		}
	}
	else if (c) /* no, so do we have a condition in slot? */
	{
		/* yes! */
		return EvaluateCondition(c, priv);
	}
	
	/* we had no condition or test, so return true. */
	return TRUE;
}

int EvaluateCondition(Condition* cond, mlfiPriv_t *priv)
{	
	int firstEval, secondEval;
	
	VerbosePrint(1, "%s", "(");
	if (cond->negate) {
		VerbosePrint(1, "%s", "!(");
	}
	
	firstEval = EvaluateConditionParameter(cond->p1, cond->t1, priv, cond->negate);
	if (cond->negate) {
		firstEval = !firstEval;
		if (firstEval) {
			VerbosePrint(1, "%s ", ") [true]");
		} else {
			VerbosePrint(1, "%s ", ") [false]");
		}
	}
	if (cond->type == kConditionTypeUndefined)
	{
		VerbosePrint(1, "%s", ")");
		return firstEval;
	}
	else if (cond->type == kConditionTypeAnd)
	{
		if (!firstEval)
		{
			VerbosePrint(1, "%s", ")");
			return FALSE;
		}
	}
	else /* kConditionTypeOr */
	{
		if (firstEval)
		{
			VerbosePrint(1, "%s", ")");
			return TRUE;
		}
	}
	
	secondEval = EvaluateConditionParameter(cond->p2, cond->t2, priv, FALSE);
	
	VerbosePrint(1, "%s", ")");

	return secondEval;
	/* this is sort of weird, but here's the deal. If it's an AND we already
		know the first branch was true because otherwise we would've short
		circuited out by now. If it's an OR then we already know that the
		first condition was false, because otherwise we'd have short circuited
		out by now. If it's an Undefined then it can't get here either because
		it always drops out after the first condition.
		
		So the end result is that it's either
		TRUE && secondEval    or
		FALSE || secondEval
		both of which are functionally == to secondEval. */
}

void DeleteCondition(Condition* c)
{
	if (c->t1)
		free(c->t1);
	else if (c->p1)
		DeleteCondition(c->p1);
	if (c->t2)
		free(c->t2);
	else if (c->p2)
		DeleteCondition(c->p2);
	free(c);
}


Condition* ParseCondition(FILE* file)
{
	/*
		Parse a condition. We don't really know what the condition's deal is, just
		that it's a condition. When we've parsed it we return the condition.
		
		The first thing we should encounter is a Test or a bracket.
		
		If we find a bracket, we need to call ParseCondition() again from this
		point. 
	*/
	
	Condition	*cond;
	char		input;
	
	cond = malloc(sizeof(Condition));
	
	InitialiseCondition(cond);
	
	input = EatWhitespace(file);
	if (input == '!') /* negation */
	{
		/* turn on negation flag */
		cond->negate = TRUE;
		/* get the next interesting character */
		input = EatWhitespace(file);
		
		/* We only have one negation flag. In a nutshell, the reason for this
		   is that in our implementation the interesting stuff is always on
		   the left branch of something, and that left branch is what the
		   negation is applying to. */
	}
	switch (input)
	{
		case ' ':
		case '\t':
		case '\v':
		case '\n':
		case '\r':
			/* whitespace - ignore */
			break;
		case '\0':
			/* unexpected EOF - panic */
			fprintf(stderr, "Parsing error - unexpected End of File\n");
			exit(EX_SOFTWARE);
		case 'S':
		case 's':
			/* S is hopefully the start of sender */
			CheckStringIsNextInFile(file, "ender", "Parsing error - expected Sender\n");
			cond->t1 = ParseTest(file, kTestSMTPSender);
			break;
		case 'F':
		case 'f':
			/* F is hopefully "from" */
			CheckStringIsNextInFile(file, "rom", "Parsing error - expected From\n");
			cond->t1 = ParseTest(file, kTestFromAddress);
			break;
		case 'T':
		case 't':
			/* T is hopefully "to" */
			CheckStringIsNextInFile(file, "o", "Parsing error - expected To\n");
			cond->t1 = ParseTest(file, kTestToAddress);
			break;
		case '(':
			cond->p1 = ParseCondition(file);
			break;
		case 'N':
		case 'n':
			/* N is hopefully "not matched" */
			CheckStringIsNextInFile(file, "ot matched", "Parsing error - expected Not Matched\n");
			cond->t1 = ParseTest(file, kTestNoMatch);
			break;
		case 'M':
		case 'm':
			/* M is hopefully "match all */
			CheckStringIsNextInFile(file, "atch all", "Parsing error - expected Match All\n");
			cond->t1 = ParseTest(file, kTestMatchAll);
			break;
		default:
			fprintf(stderr, "Parsing error - unexpected character %c\n", input);
			exit(EX_SOFTWARE);
			break;
	}
	
	input = EatWhitespace(file);
	
	switch (input)
	{
		case 'O':
		case 'o':
			/* OR hopefully */
			CheckStringIsNextInFile(file, "r", "Parsing error - expected OR\n");
			/* if we made it this far then it was OR, which means we have to start a new condition */
			cond->type = kConditionTypeOr;
			cond->p2 = ParseCondition(file);
			break;
		case 'a':
		case 'A':
			CheckStringIsNextInFile(file, "nd", "Parsing error - expected AND\n");
			cond->type = kConditionTypeAnd;
			cond->p2 = ParseCondition(file);
			break;
		case ')':
			/* we just got to the end of a bracket, so we can just return out. */
			return cond;
		case '}':
			/* this should be the end of the rule, so we return out */
			return cond;
		default:
			/* uncovered case, which means a fatal error */
			fprintf(stderr, "Parsing error - expected o, a or ), found \"%c\"\n", input);
			break;
	}
	return cond;
}
